/*
 * General ITS-like input editor.
 * This program runs under 4.2BSD UNIX.
 * Copyright 1989 Leigh L. Klotz, Jr.
 * KLOTZ@AI.MIT.EDU
 *
 *
 * Documentation:
 * It's like ITS.
 * 
 * Further documentation:
 * Run this program with arg1 being a program name and the rest of the
 * args being the args to the program.  It will let you edit a buffer
 * of text, and then on EOF will start up the program with the args
 * and send it the buffer of text.  The input editor lets you rub out
 * past the beginning of the line, all the way back to the beginning
 * of the buffer.  The word erase command works like meta-rubout in EMACS.
 *
 * Further features:
 * If stdin is not a tty, then this program skips the input editing and
 * just passes off all the input to the named program.  That way you
 * can just alias things like "mail" to be "iedit mail" and still get
 * the normal behavior with things like "mail foo < bar."
 *
 *
 * Examples:
 *
 * % iedit mail klotz@ai.ai.mit.edu
 *   type your text here
 * ^D
 * 
 * Bugs:
 * . Should get the editing characters from the tty description or wherever.
 * . Maybe should use curses or termcap or terminfo, but not if it gets too huge.
 * . Shouldn't use realloc -- the documentation says it can trash the buffer.
 * . Any way we can avoid losing the first few characters on slow machines?
 * . Some errors might still cause tty params to lose.
 */

#include <signal.h>
#include <sgtty.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/param.h>

extern int errno;
char *getenv();
char *malloc(), *realloc();

#define yes (1)
#define no (0)

#define BACKSPACE	010		/* backspace character       ^H */
#define DELCHAR         0177            /* delete character             */
#define WORDERASE	027		/* word erase character      ^W */
#define BUFRTYPE	014		/* buffer retype	     ^L */
#define LINEERASE	025		/* line kill character       ^U */
#define LINERTYPE	022		/* line reprint character    ^R */
#define USEREOF         004             /* input editor eof char     ^D */

#define INITIAL_TEXT_BUFFER_SIZE 8192
#define BUFFER_GROWTH_SIZE  1024

struct sgttyb sgttyb;			/* tty modes when interactive	*/
struct sgttyb rsgttyb;			/* original tty modes		*/

FILE *outstream;

char *text_buffer;
char *buffer_end_ptr;

void interact();
void error_write();
void buffer_retype();
void line_retype();

main(argc, argv)
int argc;
char **argv;
{
  char *cmdbuf;

  if (argc < 2) {
    fputs("usage ", stderr);
    fputs(argv[0], stderr);
    fputs(": program [args ...]\n", stderr);
    exit(-1);
  }

  /* Copy the args to a buffer to pass to POPEN */
  {
    int argno;
    unsigned space_required;
    char *p;

    /* figure the required space for the command buffer */
    space_required = 0;
    for(argno = 1, space_required = 0; argno < argc; argno++) {
      space_required += strlen(argv[argno]) + 1;
    }
    space_required++;				/* space for trailing \0 */
    /* Allocate the space */
    cmdbuf = malloc(space_required);
    if(cmdbuf == NULL) {
      fputs("error: could not allocate space for command line", stderr);
      goto error_exit;
    }

    /* Put the args into the buffer */
    for (p = cmdbuf, argno = 1; argno < argc; argno++) {
      strcpy(p, argv[argno]);
      p += strlen(argv[argno]);
      *p++ = ' ';
    }
    *p = '\0';
  }

  /* If we are not reading from a TTY, just pass off directly to the program
   * and don't do any input editing. */
  if (!isatty(fileno(stdin)))  {
    outstream = popen(cmdbuf, "w");
    if (outstream == 0) goto error_exit;
    while(yes) {
      char buffer[256];

      int nitems;
      nitems = fread(buffer, sizeof(char), sizeof buffer, stdin);
      if (nitems != 0) {
	nitems = fwrite(buffer, sizeof(char), nitems, outstream);
	if (nitems == 0) goto error_exit;
      }
      else break;
    }
  }
  else {
    setup();
    interact();
    if (text_buffer == NULL) {
      fputs("error: could not allocate space for message text", stderr);
      goto error_exit;
    }
    outstream = popen(cmdbuf, "w");
    if (outstream == 0) { error_write(); goto error_exit; }
    fwrite(text_buffer, sizeof(char), buffer_end_ptr-text_buffer, outstream);
    errno = pclose(outstream);
    if (errno != 0) { perror(cmdbuf); error_write(); goto error_exit_1; }
  }
  reset_tty_params();
  exit(0);

error_exit:
  if (errno != 0) perror("iedit");
error_exit_1:
  reset_tty_params();
  exit(errno);
}

void error_write()
{
  char fnbuf[MAXPATHLEN];
  char *home, *p;

  if (text_buffer != NULL) {
    home = getenv("HOME");
    if (home != 0) {
      strcpy(fnbuf, home);
      p = fnbuf + strlen(home);
      if (strcmp(fnbuf, "/") != 0)
	*p++ = '/';
    } else p = fnbuf;
    strcpy(p, "dead.iedit");
    outstream = fopen(fnbuf, "w");
    if (outstream > 0) {
      fputs("Saving text to ", stderr);
      fputs(fnbuf, stderr);
      fputc('\n', stderr);
      fwrite(text_buffer, sizeof(char), buffer_end_ptr-text_buffer, outstream);
      fclose(outstream);
    } else {
      if (errno != 0) perror("Error: Can't save text");
    }
  }
}


/*
 * setup - turns on CBREAK, turns off ECHO.  Also trap signals.
 * Allocates buffer.
 * Read any typeahead into buffer.
 */
setup()
{
  extern int signal_byebye();
#ifndef hpux
  extern int suspend();
#endif

  ioctl(0, TIOCGETP, &sgttyb);

  rsgttyb = sgttyb;

  signal(SIGINT, signal_byebye);
  signal(SIGQUIT, signal_byebye);
#ifndef hpux
  signal(SIGTSTP, suspend);
#endif

#ifndef hpux
  sgttyb.sg_flags |= CBREAK;
  sgttyb.sg_flags &= ~ECHO;
  ioctl(0, TIOCSETP, &sgttyb);
#endif

  text_buffer = malloc(INITIAL_TEXT_BUFFER_SIZE);
  if(text_buffer == NULL) {
    fputs("error: could not allocate space for message text", stderr);
    exit(-1);
  }
}

void display_continued_prompt()
{
  write(1, "...", 3);
}

void terpri()
{
  write(1, "\n", 1);
}

void display_prompt()
{
  write(1, "Text:\n", 6);
}

void display_rubout()
{
  write(1, "\b \b", 3);
}

/*
 * interact - interact with the user.
 */
void interact()
{
  char *physical_buffer_end;
  unsigned buffer_length;

  buffer_length = INITIAL_TEXT_BUFFER_SIZE;
  physical_buffer_end = text_buffer + buffer_length;

  display_prompt();

  /*
   * Forever... read characters.  We
   * break out of this from inside.
   */
  buffer_end_ptr = text_buffer;
  while (yes) {
    char c;

    if (read(0, &c, 1) <= 0) goto send;

    switch (c) {
    case BACKSPACE:
    case DELCHAR:
      /*
       * If not at the beginning of a line,
       * back up one character.
       */
      if (buffer_end_ptr == text_buffer) continue;
      if (buffer_end_ptr[-1] != '\n')  {
	display_rubout();
	buffer_end_ptr--;
      } else {
	buffer_end_ptr--;
	buffer_retype();
      }
      continue;

    case WORDERASE:
      /*
       * Until we hit beginning of line
       * or beginning of word, back up.
       */
      {
	int redisp_flag = no;
	
	while (!isalpha(buffer_end_ptr[-1])) {
	  if (buffer_end_ptr[-1] == '\n') {
	    buffer_end_ptr--;
	    redisp_flag = yes;
	  } else {
	    if (redisp_flag == no)
	      display_rubout();
	    buffer_end_ptr--;
	  }
	}
	while (isalpha(buffer_end_ptr[-1])) {
	  if (buffer_end_ptr[-1] == '\n') {
	    buffer_end_ptr--;
	    redisp_flag = yes;
	  } else {
	    if (redisp_flag == no)
	      display_rubout();
	    buffer_end_ptr--;
	  }
	}
	if (redisp_flag) buffer_retype();
      }
      continue;

    case LINEERASE:
      /*
       * Until we hit beginning of line,
       * back up.
       */
      {
	int redisplay_flag = no;

	if (buffer_end_ptr[-1] == '\n') {
	  redisplay_flag = yes;
	  buffer_end_ptr--;
	}
	while (buffer_end_ptr[-1] != '\n' && buffer_end_ptr > text_buffer) {
	  if (!redisplay_flag)
	    display_rubout();
	  buffer_end_ptr--;
	}
	if (redisplay_flag) buffer_retype();
      }
      continue;

    case BUFRTYPE:
      buffer_retype();
      continue;

    case LINERTYPE:
      /*
       * Retype the line
       */
      line_retype();
      continue;

    case USEREOF:
      goto send;

    default:
      /*
       * Allocate space if necessary.
       */
      if (buffer_end_ptr == physical_buffer_end) {
	char *newbuf;
	newbuf = realloc(text_buffer, buffer_length + BUFFER_GROWTH_SIZE);
	if (newbuf == NULL) {
	  fputs("\nCouldn't make message buffer longer\n", stderr);
	}
	else {
#ifdef REALLOC_DEBUG
	  fputs("\007", stderr);
#endif
	  buffer_end_ptr = newbuf + (buffer_end_ptr - text_buffer);
	  text_buffer = newbuf;
	  buffer_length += BUFFER_GROWTH_SIZE;
	  physical_buffer_end = text_buffer + buffer_length;
	}
      }
      /*
       * Echo the character and copy it.
       */
      if (buffer_end_ptr < physical_buffer_end) {
	write(1, &c, 1);
	*buffer_end_ptr++ = c;
	continue;
      }
    }
  }
 send:
  return;
}

void buffer_retype()
{
#ifdef HPUX
  /* is this "right" */
  system("/usr/bin/clear")
#else
  system("/usr/ucb/clear");
#endif
  /*
   * Retype the buffer.
   */
  display_prompt();
  write(1, text_buffer, buffer_end_ptr-text_buffer);
}

void line_retype()
{
  char *p;

  terpri();
  for (p = buffer_end_ptr-1; p > text_buffer; p--) 
    if (*p == '\n') break;
  write(1, p+1, buffer_end_ptr-p-1);
}


/*
 * byebye - called on signal exit.
 */
signal_byebye()
{
  terpri();
  error_write();
  reset_tty_params();
  exit(0);
}

reset_tty_params()
{
/* why is this ifndef hpux? */
#ifndef hpux
  ioctl(0, TIOCSETP, &rsgttyb);
#endif
}



#ifndef hpux
/*
 * suspend - reset tty modes and supend ourselves.
 */
suspend()
{
	long blocked;
	extern int suspend();

	/*
	 * Reset tty modes and suspend.
	 */
	ioctl(0, TIOCSETP, &rsgttyb);
	signal(SIGTSTP, SIG_DFL);
	blocked = sigsetmask(0);
	kill(0, SIGTSTP);

	/*
	 * We come here on SIGCONT.  Reset
	 * the signal mask and tty modes.
	 */
	sigsetmask(blocked);
	signal(SIGTSTP, suspend);
	ioctl(0, TIOCSETP, &sgttyb);	
	display_continued_prompt();
	line_retype();
}
#endif

